/*
 * Simple C code to measure udp round trip times
 *
 * (C) 2009 Linux Foundation
 *		Christoph Lameter <cl@liuux-foundation.org>
 */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <pthread.h>
#include <linux/types.h>

#define MAGIC 0xFFEEDDCC

struct msg {
	int magic;
	int ad;
	int port;
	__u64 timestamp;
	char payload[10];
};

void fatal(const char *x, ...)
{
	va_list ap;

	va_start(ap, x);
	vfprintf(stderr, x, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

static inline long min(int a, int b)
{
	if (a < b)
		return a;
	return b;
}

static inline __u64 get_timestamp(void)
{
	struct timespec spec;

	clock_gettime(CLOCK_REALTIME, &spec);
	return spec.tv_nsec + spec.tv_sec * 1000000000L;
}

int port = 9000;	/* Port to listen to or send to */
int samples = 100000;	/* Number of samples to take */
int runtime = 30;		/* Maximum runtime */
int ttl = -1;		/* Set ttl */
int length = 300;

int origport = 9001;	/* Origination port for ping */
struct sockaddr_in bindaddr;	/* IP address to bind to */
struct sockaddr_in addr;

int details = 0;	/* Print detailed statistics */
int verbose = 0;
int stream = 0;		/* Run a streaming test not ping pong */
int rate = 0;		/* Limit rate */
struct timespec wait_time = { 0, 0 };		/* Pause between packets */

int inblocksize = 10000;	/* Size of the block to send */
int outblocksize = 1000;	/* Size of the block to send */

int yes = 1;

int sock;

int open_sock(struct sockaddr_in *sa)
{
	int err;
	struct sockaddr inad;

	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) {
		perror("socket");
		return -2;
	}

	err = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
	if (err) {
		perror("reuseaddr");
		return -3;
	}

	((struct sockaddr_in *)&inad)->sin_family = AF_INET;
	((struct sockaddr_in *)&inad)->sin_addr.s_addr = bindaddr.sin_addr.s_addr;
	((struct sockaddr_in *)&inad)->sin_port = bindaddr.sin_port;
	if (bind(sock, &inad, sizeof(inad))) {
		perror("bind");
		return -4;
       	}

	if (ttl >= 0) {
		err = setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
		if (err)
			perror("setsockopt IP_MULTICAST_TTL\n");
		return -5;
	}
	return sock;
}

enum sender_mode { SENDER_OFF, SENDER_IDLE, SENDER_ACTIVE };

volatile enum sender_mode mode = SENDER_OFF;


#define BUFFER_SIZE 1000000

struct sockaddr sendaddr;
unsigned sendaddrlen;

char buffer[BUFFER_SIZE];
volatile char *buffer_sent = buffer;	/* Part of the buffer that was alrady sent */
volatile char *buffer_end = buffer;

void *echo_sender_thread(void *private)
{
	int rc;

	mode = SENDER_IDLE;
	while (1) {
		char *end = (char *)buffer_end;
		char *sent = (char *)buffer_sent;

		if (end == sent) {
			/* spin idle */
			continue;
		}

		/* Send as much as we can */
		mode = SENDER_ACTIVE;
		if (end < sent) {
			/* Wraparound */
			rc = sendto(sock, sent, buffer + BUFFER_SIZE - sent,
						0, &sendaddr, sendaddrlen);
			if (rc< 0) {
				perror("sendto");
				fatal("sender: error on sendto");
			}
			buffer_sent = sent = buffer;
		}

		rc = sendto(sock, sent, min(outblocksize, end - sent),
						0, &sendaddr, sendaddrlen);
		if (rc< 0) {
			perror("sendto");
			fatal("sender: error on sendto");
		}
		buffer_sent += rc;
		mode = SENDER_IDLE;
	}
}

char msg[10000];


int udp_echo(struct sockaddr_in *sa)
{
	int err;
	int nbytes;
	struct sockaddr inad;
	unsigned inadlen = sizeof(inad);


	open_sock(sa);
	if (sock < 0)
		return err;

	printf("Waiting for udp traffic on port %d ...\n", ntohs(sa->sin_port));
	sendaddrlen = sizeof(sendaddr);
	nbytes = recvfrom(sock, msg, sizeof(msg), MSG_PEEK,
				(struct sockaddr *)&sendaddr, &sendaddrlen);
	if (nbytes < 0) {
		perror("recfrom");
                printf("UDP receive failed %d\n", nbytes);
                return 6;
        }

	printf("Msg Peek resulted in receiving %d bytes. Starting echo\n", nbytes);

	if (stream) {
		pthread_attr_t attr;
		pthread_t sender_thread_t;
		int rc;
		int i;

		/* Buffer must be present to avoid page faults so touch it */
		for (i = 0; i < BUFFER_SIZE; i += 4096)
			buffer[i] = 1;

		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		rc = pthread_create(&sender_thread_t, &attr, echo_sender_thread, sa);
		if (rc < 0) {
			perror("pthread create");
			fatal("Cannot create sender thread\n");
		}
		/* Wait for sender to become active */
		while (mode == SENDER_OFF) ;

		while (1) {
			char *end = (char *)buffer_end;


			nbytes = recvfrom(sock, end, min(inblocksize, buffer + BUFFER_SIZE - end),
							0, &inad, &inadlen);
			if (nbytes < 0) {
				perror("recvfrom");
				break;
			}
			buffer_end += nbytes;
			if (buffer_end  == buffer + BUFFER_SIZE)
				buffer_end = buffer;
		}
	} else {
		/* Ping pong mode */
		while(1) {
			nbytes = recvfrom(sock, msg, min(inblocksize, sizeof(msg)),
								0, &inad, &inadlen);
			if (nbytes < 0) {
				perror("recvfrom");
				break;
			}
			if (sendto(sock, msg, nbytes, 0, &inad, inadlen) < 0) {
				perror("sendto");
				break;
			}
		}
	}
	close(sock);
	return 0;
}

struct msg outmsg;

void *ping_sender_thread(void *private)
{
	int i;

	mode = SENDER_ACTIVE;
	for(i = 1; i < samples; i++) {
		outmsg.timestamp = get_timestamp();

		if (rate) {
			struct timespec tmp;

			nanosleep(&wait_time, &tmp);
		}

		if (sendto(sock, &outmsg, length, 0,
				&sendaddr, sizeof(sendaddr)) < 0) {
			perror("sendto");
			break;
		}
	}
	mode = SENDER_OFF;
	/* Termination packet */
	sendto(sock, &outmsg, sizeof(outmsg.magic), 0, &sendaddr, sizeof(sendaddr));
	return NULL;
}

int udp_ping(struct sockaddr_in *sa)
{
	int nbytes;
	struct msg *inmsg;
	int i, max, min;
	long long sum  =0;
	char *x;
	int sample = 0;
	int bad = 0;
	int *tdiff;

	open_sock(sa);

	printf("udp ping %s:%d -> %s:%d %d samples ....\n",
				 strdupa(inet_ntoa(bindaddr.sin_addr)), origport,
				 strdupa(inet_ntoa(sa->sin_addr)), port,
				 samples);
	outmsg.magic = MAGIC;
	outmsg.ad = sa->sin_addr.s_addr;
	outmsg.port = htons(origport);

	tdiff = malloc(samples * sizeof(*tdiff));
	for (i = 0; i < 10; i++)
		outmsg.payload[i] = i;

	if (stream) {
		pthread_attr_t attr;
		pthread_t sender_thread_t;
		int rc;

		sendaddr = *(struct sockaddr *)sa;
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		rc = pthread_create(&sender_thread_t, &attr, ping_sender_thread, sa);
		if (rc < 0) {
			perror("pthread create");
			fatal("Cannot create sender thread\n");
		}
		/* Wait for sender to become active */
		while (mode == SENDER_OFF) ;
	}
	for (i = 0; i < samples; i++) {
		long diff;

		if (!stream) {
			outmsg.timestamp = get_timestamp();

			if (rate) {
				struct timespec tmp;

				nanosleep(&wait_time, &tmp);
			}

			if (sendto(sock, &outmsg, length, 0, &addr, sizeof(addr)) < 0) {
				perror("sendto");
				break;
			}
		} else {
			if (mode == SENDER_OFF)
				break;
		}

		nbytes = recvfrom(sock, &msg, length, 0, NULL, NULL);
		if (nbytes < 0) {
			perror("recvfrom");
			break;
		}
		for (inmsg = (void *)msg; nbytes > 0; inmsg++, nbytes -= length) {
			if (inmsg->magic != MAGIC) {
				x = (void *)inmsg;
				printf("%d. Reply format error lenght=%d %x %x %x %x %x %x %x %x %x %x.\n",
				i, nbytes, x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9]);
			}
			if (nbytes == sizeof(inmsg->magic))
				/* End packet */
				break;

			diff = get_timestamp() - inmsg->timestamp;
			if (diff <= 0 || diff > 1000000000) {
				/* Bad sample. Skip */
				bad++;
				if (verbose)
					printf("Iteration %d. %.2fus bytes=%d\n", i, diff / 1000.0, nbytes);
			} else
				tdiff[sample++] = diff;
		}
	}
	min = tdiff[0]; max = 0; sum = 0;
	for (i = 0; i < sample; i++) {
		int x = tdiff[i];

		sum += x;
		if (x < min)
			min = x;
		if (x > max)
			max = x;
	}

	if (sample) {
		printf("%.2fus (%.2fus/%.2fus/%.2fus)", sum / 1000.0,
			 min / 1000.0, sum / sample / 1000.0, max / 1000.0);
		if (bad)
			printf(" %d bad samples", bad);
	} else
		printf("Nothing received back\n");

	printf("\n");
	if (details && sample) {
		int *ns;

		ns = malloc((max / 1000 + 1) * sizeof(int));
		for (i = 0; i < samples; i++)
			ns[tdiff[i] / 1000]++;

		printf("\nusecs percent  samples\n----------------------\n");
		for (i = 1; i < max / 1000 + 1; i++) {
			if (ns[i] > sample / 1000)
				printf("%5d %7.1f %8d\n", i,
					ns[i] * 100.0 / sample, ns[i]);
		}
	}
	return 0;
}

void usage(void)
{
	printf("udpping 8/5/2009 UDP ping pong measuring tool\n"
		"(C) 2009 Linux Foundation, Christoph Lameter <cl@linux-foundation.org>\n\n"
		"udpping [destination]\n"
		"-p<port>|--port <port>            Change the port to use [9000]\n"
		"-n<samples>|--number <samples>    Specify the number of samples [100000]\n"
		"-t<secs>|--time <secs>            Maximum run time of ping[30]\n"
		"-l<length>|--length <length>      Set length of packet\n"
		"-g<port>|--origport <port>        Set origination port [9001]\n"
		"-b<ip>|--bind <ip>                Bind to a certain IP address\n"
		"-d|--details                      Show detailed distribution of response times\n"
		"-s|--stream                       Do a stream test not a ping pong test\n"
		"-h|--help                         Print usage information\n"
		"-i<bytes>|--inblocksize <bytes>   Maximum number of bytes to accept with one syscall\n"
		"-o<bytes>|--outblocksize <bytes>  Maximum number of bytes to send with one syscall\n"
		"-r<packets>|--rate <packets>      Packets to send per second.\n"
	);
}

struct option opts[] = {
	{ "port", 1, NULL, 'p' },
	{ "number", 1, NULL, 'n' },
	{ "time", 1, NULL, 't' },
	{ "length", 1, NULL, 'l' },
	{ "origport", 1, NULL, 'g' },
	{ "bind", 1, NULL, 'b' },
	{ "details", 0, NULL, 'd' },
	{ "help", 0, NULL, 'h' },
	{ "verbose", 0, NULL, 'v' },
	{ "stream", 0, NULL, 's' },
	{ "inblocksize", 0, NULL, 'i' },
	{ "outblocksize", 0, NULL, 'o' },
	{ "rate", 0, NULL, 'r' },
	{ NULL, 0, NULL, 0 }
};

int main(int argc, char *argv[])
{
	int c;

	addr.sin_family = AF_INET;
	bindaddr.sin_family = AF_INET;
	bindaddr.sin_addr.s_addr = htonl(INADDR_ANY);

	while ((c = getopt_long(argc, argv, "vhedsp:n:t:l:o:b:i:o:r:",
							opts, NULL)) != -1)
	switch (c) {
	case 'h' :
		usage();
		return 0;

	case 'd' :
		details = 1;
		break;

	case 's' :
		stream = 1;
		break;

	case 'p' :
		port = atoi(optarg);
		if (port < 1024 || port > 60000)
			fatal("Invalid port number %d\n", port);
		break;

	case 'g' :
		origport = atoi(optarg);
		if (origport < 1024 || origport > 60000)
			fatal("Invalid port number %d\n", origport);
		break;

	case 'n' :
		samples = atoi(optarg);
		if (samples < 10 || samples > 1000000000)
			fatal("Invalid number of samples %d\n", samples);
		break;

	case 't' :
		runtime = atoi(optarg);
		if (runtime < 1 || runtime > 360)
			fatal("Invalid duration %d seconds\n", runtime);
		break;

	case 'l' :
		length = atoi(optarg);
		if (length < sizeof(struct msg) - 10 || length > sizeof(msg))
			fatal("Invalid packet length %d\n", length);
		break;

	case 'b' :
		if (!inet_aton(optarg, &bindaddr.sin_addr))
			printf("Invalid source host address\n");
		break;

	case 'v':
		verbose++;
		break;

	case 'i' :
		inblocksize = atoi(optarg);
		if (inblocksize < 2 || inblocksize > 1000000)
			fatal("Invalid inblock size %d\n", inblocksize);
		break;

	case 'o' :
		outblocksize = atoi(optarg);
		if (outblocksize < 2 || outblocksize > 60000)
			fatal("Invalid outblocksize number %d\n", outblocksize);
		break;

	case 'r' :
		rate = atoi(optarg);
		if (rate > 100000 || rate < 10)
			fatal("Invalid rate %d\n", rate);

		wait_time.tv_nsec = 1000000000 / rate;
		break;

	default:
		fatal("%s: Invalid option '%c'\n", argv[0], optopt);
	}

	if (argc > 1 + optind) {
		usage();
		return 1;
	}

	if (argc == 1 + optind) {
		if (!inet_aton(argv[optind], &addr.sin_addr))
			fatal("Invalid destination host address\n");

		addr.sin_port = htons(port);
		bindaddr.sin_port = htons(origport);
		return udp_ping(&addr);
	} else {
		bindaddr.sin_port = htons(port);
		return udp_echo(&bindaddr);
	}
}

